# -*- coding: utf-8 -*-
from collections import namedtuple
from watson.common.imports import get_qualified_name

ListenerDefinition = namedtuple('Listener', 'callback priority only_once')


class Listener(list):

    """A list of listeners to be used in an EventDispatcher.

    A Listener Collection is a list of callbacks that are to be triggered
    by an event dispatcher. Each item in the list contains the callback, a
    priority, and whether or not the callback should only be triggered once.
    """

    require_sort = False

    def add(self, callback, priority=1, only_once=False):
        """Adds a new callback to the collection.

        Args:
            callback (callable): the function to be triggered
            priority (int): how important the callback is in relation to others
            only_once (bool): the callback should only be fired once and then removed

        Raises:
            TypeError if non-callable is added.
        """
        if not hasattr(callback, '__call__'):
            name = get_qualified_name(callback)
            raise TypeError('Callback {0} must be callable.'.format(name))
        self.append(ListenerDefinition(callback, priority, only_once))
        self.require_sort = True
        return self

    def remove(self, callback):
        """Removes all callbacks matching `callback` from the collection.

        Args:
            callback (callable): the callback to be removed.
        """
        for listener in list(self):
            if listener.callback == callback:
                super(Listener, self).remove(listener)

    def sort_priority(self):
        """Sort the collection based on the priority of the callbacks.
        """
        if self.require_sort:
            self.sort(key=lambda listener: listener.priority, reverse=True)
            self.require_sort = False

    def __contains__(self, callback):
        return [listener for listener in self if listener.callback == callback]

    def __repr__(self):
        return (
            '<{0} callbacks:{1}>'.format(get_qualified_name(self), len(self))
        )


class Result(list):

    """A list of responses from a EventDispatcher.trigger call.

    A result collection contains all the resulting output from an event that has
    been triggered from an event dispatcher. It provides some convenience methods
    to deal with the results.
    """

    def first(self, default=None):
        """Return the first result from the list.

        Args:
            default (mixed): The value to return if the index doesn't exist

        Returns:
            mixed: The first result
        """
        try:
            return self[0]
        except IndexError:
            return default

    def last(self, default=None):
        """Return the last result from the list.

        Args:
            default (mixed): The value to return if the index doesn't exist

        Returns:
            mixed: The first result
        """
        try:
            return self[-1]
        except IndexError:
            return default

    def __repr__(self):
        return '<{0} results:{1}>'.format(get_qualified_name(self), len(self))



# -*- coding: utf-8 -*-
from watson.common.imports import get_qualified_name


class Event(object):

    """A base event that can be subclassed for use with an EventDispatcher.

    Example:

    .. code-block:: python

        def my_listener(event):
            print(event.params['config'])

        dispatcher.add('MyEvent', my_listener)

        event = Event('MyEvent')
        event.params['config'] = {'some': 'config'}
        dispatcher.trigger(event)
    """
    name = None
    _stop_propagation = False
    params = None
    target = None

    def __init__(self, name, target=None, params=None):
        """Initializes the event.

        Initialize the Event based on an event name. The name will be used
        when the event is triggered from the event dispatcher.

        Args:
            name (string): the name of the event
            target (mixed): the originating target of the event
            params (dict): the params associated with the event
        """
        self.name = name
        self.target = target
        self.params = params or {}

    @property
    def stopped(self):
        """Return whether or not the event has been stopped.
        """
        return self._stop_propagation

    def stop_propagation(self):
        """Prevents the event from triggering any more event listeners.

        This should be used within an event listener when you wish to halt
        any further listeners from being triggered.
        """
        self._stop_propagation = True

    def __repr__(self):
        return '<{0} name:{1}>'.format(get_qualified_name(self), self.name)
